from typing import (
    Optional,
    List,
    Union,
    Protocol,
    runtime_checkable,
    Dict,
    Literal,
)
from pydantic import Field, field_validator
from koil.composition import KoiledModel
from datetime import timedelta
from .compose_spec import ComposeSpec
import json
import os
from dokker.errors import DokkerError
from dokker.types import ValidPath, LogStream
from dokker.command import astream_command


class CLIError(DokkerError):
    """An error that is raised when the CLI fails to run."""

    pass


@runtime_checkable
class CLIBearer(Protocol):
    """A CLIBearer is an object that has a CLI.

    This is a protocol that can be used to type hint objects that have a CLI.
    """

    async def aget_cli(self) -> "CLI":
        """Returns the CLI for the object."""
        ...


class CLI(KoiledModel):
    """A CLI object that represents the docker-compose CLI.

    This is a pydantic model that can be used to build the docker-compose CLI
    command. It also contains methods for running the CLI command
    asynchronously.

    """

    config: Optional[ValidPath] = None
    context: Optional[str] = None
    debug: Optional[bool] = None
    host: Optional[str] = None
    log_level: Optional[str] = None
    tls: Optional[bool] = None
    tlscacert: Optional[ValidPath] = None
    tlscert: Optional[ValidPath] = None
    tlskey: Optional[ValidPath] = None
    tlsverify: Optional[bool] = None
    compose_files: List[ValidPath] = Field(default_factory=lambda: ["docker-compose.yml"])
    compose_profiles: List[ValidPath] = Field(default_factory=list)
    compose_env_file: Optional[ValidPath] = Field(default=".env")
    compose_project_name: Optional[str] = None
    compose_project_directory: Optional[ValidPath] = None
    compose_compatibility: Optional[bool] = None
    client_call: List[str] = Field(default_factory=lambda: ["docker", "compose"])

    @field_validator("compose_files")
    def _validate_compose_files(cls, v: str) -> list[ValidPath]:
        x: list[ValidPath] = []
        for vo in v:
            if os.path.exists(vo):
                x.append(vo)
            else:
                raise ValueError(f"Compose File {v} does not exist.")

        return x

    @property
    def docker_cmd(self) -> List[str]:
        """Builds the docker command. This is the base prepended
        command that will be run by the CLI.
        """
        result = self.client_call

        if self.compose_files:
            for compose_file in self.compose_files:
                result += ["--file", str(compose_file)]

        if self.config is not None:
            result += ["--config", str(self.config)]

        if self.context is not None:
            result += ["--context", self.context]

        if self.debug:
            result.append("--debug")

        if self.host is not None:
            result += ["--host", self.host]

        if self.log_level is not None:
            result += ["--log-level", self.log_level]

        if self.tls:
            result.append("--tls")

        if self.tlscacert is not None:
            result += ["--tlscacert", str(self.tlscacert)]

        if self.tlscert is not None:
            result += ["--tlscert", str(self.tlscert)]

        if self.tlskey is not None:
            result += ["--tlskey", str(self.tlskey)]

        if self.tlsverify:
            result.append("--tlsverify")

        return result

    async def astream_docker_logs(
        self,
        tail: Optional[str] = None,
        follow: bool = False,
        no_log_prefix: bool = False,
        timestamps: bool = False,
        since: Optional[str] = None,
        until: Optional[str] = None,
        services: Union[str, List[str]] = [],
    ) -> LogStream:
        """Runs the docker logs command asynchronously."""
        full_cmd = self.docker_cmd + ["logs", "--no-color"]
        if tail is not None:
            full_cmd += ["--tail", tail]
        if follow:
            full_cmd.append("--follow")
        if no_log_prefix:
            full_cmd.append("--no-log-prefix")
        if timestamps:
            full_cmd.append("--timestamps")
        if since is not None:
            full_cmd += ["--since", since]
        if until is not None:
            full_cmd += ["--until", until]

        if services:
            if isinstance(services, str):
                services = [services]
            full_cmd += services

        async for line in astream_command(full_cmd):
            yield line

    async def astream_down(
        self,
        remove_orphans: bool = False,
        remove_images: Optional[str] = None,
        timeout: Optional[int] = None,
        volumes: bool = False,
    ) -> LogStream:
        """Runs the docker-compose down command asynchronously."""
        full_cmd = self.docker_cmd + ["down"]
        if remove_orphans:
            full_cmd.append("--remove-orphans")
        if remove_images is not None:
            full_cmd.append(f"--rmi {remove_images}")
        if timeout is not None:
            full_cmd.append(f"--timeout {timeout}")
        if volumes:
            full_cmd.append("--volumes")

        async for line in astream_command(full_cmd):
            yield line

    async def astream_pull(
        self,
        services: Union[List[str], str, None] = None,
        ignore_pull_failures: bool = False,
        include_deps: bool = False,
        quiet: bool = False,
    ) -> LogStream:
        """Runs the docker-compose pull command asynchronously."""
        full_cmd = self.docker_cmd + ["pull"]
        if ignore_pull_failures:
            full_cmd.append("--ignore-pull-failures")
        if include_deps:
            full_cmd.append("--include-deps")
        if quiet:
            full_cmd.append("--quiet")

        if services:
            if isinstance(services, str):
                services = [services]
            full_cmd += services

        async for line in astream_command(full_cmd):
            yield line

    async def astream_stop(
        self,
        services: Union[str, List[str], None] = None,
        timeout: Union[int, timedelta, None] = None,
    ) -> LogStream:
        """Runs the docker-compose stop command asynchronously."""
        full_cmd = self.docker_cmd + ["stop"]
        if timeout is not None:
            if isinstance(timeout, timedelta):
                timeout = int(timeout.total_seconds())

            full_cmd.append(f"--timeout {timeout}")

        if services:
            if isinstance(services, str):
                services = [services]
            full_cmd += services

        async for line in astream_command(full_cmd):
            yield line

    async def astream_restart(
        self,
        services: Union[str, List[str], None] = None,
    ) -> LogStream:
        """Runs the docker-compose restart command asynchronously."""
        full_cmd = self.docker_cmd + ["restart"]

        if services:
            if isinstance(services, str):
                services = [services]
            full_cmd += services

        async for line in astream_command(full_cmd):
            yield line

    async def astream_up(
        self,
        services: Union[List[str], str, None] = None,
        build: bool = False,
        detach: bool = False,
        abort_on_container_exit: bool = False,
        scales: Dict[str, int] = {},
        attach_dependencies: bool = False,
        force_recreate: bool = False,
        no_recreate: bool = False,
        no_build: bool = False,
        remove_orphans: bool = False,
        renew_anon_volumes: bool = False,
        no_color: bool = False,
        no_log_prefix: bool = False,
        no_start: bool = False,
        quiet: bool = False,
        wait: bool = False,
        no_attach_services: Union[List[str], str, None] = None,
        pull: Literal["always", "missing", "never", None] = None,
        stream_logs: bool = False,
    ) -> LogStream:
        """Runs the docker-compose up command asynchronously."""
        if quiet and stream_logs:
            raise ValueError("It's not possible to have stream_logs=True and quiet=True at the same time. Only one can be activated at a time.")
        full_cmd = self.docker_cmd + ["up"]
        if build:
            full_cmd.append("--build")
        if detach:
            full_cmd.append("--detach")
        if abort_on_container_exit:
            full_cmd.append("--abort-on-container-exit")
        for service, scale in scales.items():
            full_cmd.append(f"--scale {service}={scale}")
        if attach_dependencies:
            full_cmd.append("--attach-dependencies")
        if force_recreate:
            full_cmd.append("--force-recreate")
        if no_recreate:
            full_cmd.append("--no-recreate")
        if no_build:
            full_cmd.append("--no-build")
        if remove_orphans:
            full_cmd.append("--remove-orphans")
        if renew_anon_volumes:
            full_cmd.append("--renew-anon-volumes")
        if no_color:
            full_cmd.append("--no-color")
        if no_log_prefix:
            full_cmd.append("--no-log-prefix")
        if no_start:
            full_cmd.append("--no-start")
        if quiet:
            full_cmd.append("--quiet")
        if wait:
            full_cmd.append("--wait")
        if no_attach_services is not None:
            if isinstance(no_attach_services, str):
                no_attach_services = [no_attach_services]
            for service in no_attach_services:
                full_cmd.append(f"--no-attach {service}")
        if pull is not None:
            full_cmd.append(f"--pull {pull}")

        if services:
            if isinstance(services, str):
                services = [services]
            full_cmd += services

        async for line in astream_command(full_cmd):
            yield line

    async def ainspect_config(self) -> ComposeSpec:
        """Inspect the config of the docker-compose project.

        Returns
        -------
        ComposeSpec
            The compose spec of the project.

        Raises
        ------
        CLIError
            An error that is raised when the CLI fails to run.
        """
        full_cmd = self.docker_cmd + ["config", "--format", "json"]

        stdout_lines: list[str] = []

        async for source, line in astream_command(full_cmd):
            if source == "STDERR":
                continue
            elif source == "STDOUT":
                stdout_lines.append(line)
            else:
                raise ValueError(f"Unknown source: {source}")

        result = "\n".join(stdout_lines)

        try:
            return ComposeSpec(**json.loads(result))
        except Exception as e:
            raise CLIError(f"Could not inspect! Error while parsing the json: {result}") from e



from pydantic import BaseModel, ConfigDict
from typing import Callable

LogTuple = tuple[str, str]


class PrintLogger(BaseModel):
    """A logger that prints all logs to stdout"""

    model_config = ConfigDict(arbitrary_types_allowed=True)
    should_print: bool = True
    print_function: Callable[[LogTuple], None] = print

    def on_pull(self, log: LogTuple) -> None:
        """A method for logs

        Parameters
        ----------
        log : str
            The log to print
        """
        self.print_function(log)

    def on_up(self, log: LogTuple) -> None:
        """A method for logs

        Parameters
        ----------
        log : str
            The log to print
        """
        self.print_function(log)

    def on_stop(self, log: LogTuple) -> None:
        """A method for logs

        Parameters
        ----------
        log : str
            The log to print
        """
        self.print_function(log)

    def on_logs(self, log: LogTuple) -> None:
        """A method for logs

        Parameters
        ----------
        log : str
            The log to print
        """
        self.print_function(log)

    def on_down(self, log: LogTuple) -> None:
        """A method for logs

        Parameters
        ----------
        log : str
            The log to print
        """
        self.print_function(log)



from pydantic import BaseModel

LogTuple = tuple[str, str]


class VoidLogger(BaseModel):
    """A logger that omits all logs"""

    def on_pull(self, log: LogTuple) -> None:
        """A method for logs

        Parameters
        ----------
        log : str
            The log to print
        """

        pass

    def on_up(self, log: LogTuple) -> None:
        """A method for logs

        Parameters
        ----------
        log : str
            The log to print
        """
        pass

    def on_stop(self, log: LogTuple) -> None:
        """A method for logs

        Parameters
        ----------
        log : str
            The log to print
        """
        pass

    def on_logs(self, log: LogTuple) -> None:
        """A method for logs

        Parameters
        ----------
        log : str
            The log to print
        """
        pass

    def on_down(self, log: LogTuple) -> None:
        """A method for logs

        Parameters
        ----------
        log : str
            The log to print
        """
        pass



""" Dokker

Dokker is a tool for building and managing docker-compose projects.
It is designed to tightly integrate in python projects and provide
sensible defaults for common docker-compose workflows.
"""

from .deployment import Deployment, HealthCheck, Logger
from .builders import (
    mirror,
    testing,
    monitoring,
    local,
)
from .project import Project
from .projects.local import LocalProject

__all__ = [
    "Deployment",
    "HealthCheck",
    "Logger",
    "mirror",
    "testing",
    "monitoring",
    "local",
    "Project",
    "LocalProject",
]



from pydantic import BaseModel, ConfigDict, Field
from typing import Union
from pathlib import Path
from dokker.cli import CLI
from typing import Dict, Any, Optional
import logging
import os
from dokker.projects.errors import ProjectError

logger = logging.getLogger(__name__)

ValidPath = Union[str, Path]


class DokkerProjectError(ProjectError):
    """An error raised when a local project is not initialized."""

    pass


class DokkerProject(BaseModel):
    """A project that is located on the local filesystem.

    This project is a project that is located on the local filesystem.
    It can be used to run a docker-compose file locally, without
    interfering with the docker project on tear-down.
    """

    model_config = ConfigDict(arbitrary_types_allowed=True)
    base_dir: ValidPath = Field(default_factory=lambda: os.path.join(os.getcwd(), ".dokker"))
    name: str
    _outs: Optional[Dict[str, Any]] = None
    _project_dir: Optional[ValidPath] = None

    async def ainititialize(self) -> CLI:
        """A setup method for the project.

        Returns
        -------
        CLI
            The CLI to use for the project.
        """
        os.makedirs(self.base_dir, exist_ok=True)

        self._project_dir = os.path.join(self.base_dir, self.name)
        if not os.path.exists(self._project_dir):
            raise DokkerProjectError(f"No project found with the name {self.name} in {self.base_dir}. Available projects: {os.listdir(self.base_dir)}")

        compose_file = os.path.join(self._project_dir, "docker-compose.yaml")
        if not os.path.exists(compose_file):
            raise Exception("No docker-compose.yml found in the template. It appears that the template is not a valid dokker template.")

        return CLI(
            compose_files=[compose_file],
        )

    async def atear_down(self, cli: CLI) -> None:
        """Tear down the project.

        A project can implement this method to tear down the project
        when the project is torn down. This can be used to remove
        temporary files, or to remove the project from the .dokker
        directory.

        Parameters
        ----------
        cli : CLI
            The CLI that was used to run the project.

        """
        logger.info("Tearing down project, is a void operation for a local project")
        return None

    async def abefore_pull(self) -> None:
        """A setup method for the project.

        Returns:
            Optional[List[str]]: A list of logs from the setup process.
        """
        return None

    async def abefore_up(self) -> None:
        """A setup method for the project.

        Returns:
            Optional[List[str]]: A list of logs from the setup process.
        """
        return None

    async def abefore_enter(self) -> None:
        """A setup method for the project.

        Returns:
            Optional[List[str]]: A list of logs from the setup process.
        """
        return None

    async def abefore_down(self) -> None:
        """A setup method for the project.

        Returns:
            Optional[List[str]]: A list of logs from the setup process.
        """
        return None

    async def abefore_stop(self) -> None:
        """A setup method for the project.

        Returns:
            Optional[List[str]]: A list of logs from the setup process.
        """
        return None



from typing import Callable, Union, AsyncIterator, Tuple, Awaitable
from pathlib import Path

ValidPath = Union[str, Path]
LogStream = AsyncIterator[Tuple[str, str]]
LogFunction = Union[Callable[[Tuple[str, str]], Awaitable[None]], Callable[[Tuple[str, str]], None]]



from types import TracebackType
import aiohttp.client_exceptions
import aiohttp.http_exceptions
from pydantic import BaseModel, ConfigDict, Field
from typing import Dict, Optional, List, Protocol, Self, Type, runtime_checkable
from koil.composition import KoiledModel
import asyncio
from pathlib import Path
from dokker.compose_spec import ComposeSpec
from dokker.project import Project
from typing import Union
from koil import unkoil
from dokker.cli import CLI
from dokker.loggers.void import VoidLogger
from dokker.types import LogFunction
from .log_watcher import LogRoll, LogWatcher
import aiohttp
import certifi
from ssl import SSLContext
import ssl
from typing import Callable
from dokker.errors import NotInitializedError, NotInspectedError, HealthCheckError


ValidPath = Union[str, Path]


class HealthCheck(BaseModel):
    """A health check for a service.

    This class is used to check the health of a service by making a request to a given URL.
    The URL can be a string or a callable that takes the compose spec as an argument and returns a string.
    The health check will be retried a given number of times with a given timeout between retries.
    If the health check fails, an error will be raised.
    """

    model_config = ConfigDict(arbitrary_types_allowed=True)
    url: Union[str, Callable[[ComposeSpec], str]] = Field(
        description="The url to check. Can be a string or a callable that takes the compose spec as an argument and returns a string."
    )
    service: str = Field(description="The service to check.")
    max_retries: int = Field(
        default=3, description="The maximum number of retries before failing."
    )
    timeout: int = Field(default=10, description="The timeout between retries.")
    error_with_logs: bool = Field(
        default=True,
        description="Should we error with the logs of the service (will inspect container logs of the service).",
    )
    headers: Optional[Dict[str, str]] = Field(
        default_factory=lambda: {"Content-Type": "application/json"},
        description="Headers to use for the request",
    )
    ssl_context: SSLContext = Field(
        default_factory=lambda: ssl.create_default_context(cafile=certifi.where()),
        description="SSL Context to use for the request",
    )
    valid_statuses: list[int] = Field(
        default_factory=lambda: [200],
        description="The valid statuses for the health check. Defaults to 200.",
    )

    async def acheck(self, spec: ComposeSpec) -> str:
        """Check the health of the service.

        This method will make a request to the given URL and check the response status.
        If the status is not in the valid statuses, an error will be raised.
        Parameters
        ----------
        spec : ComposeSpec
            The compose spec to use for the health check.
        """
        async with aiohttp.ClientSession(
            headers=self.headers,
            connector=aiohttp.TCPConnector(ssl=self.ssl_context),
        ) as session:
            # get json from endpoint
            url = self.url if isinstance(self.url, str) else self.url(spec)

            try:
                async with session.get(url) as resp:
                    if resp.status not in self.valid_statuses:
                        raise HealthCheckError(
                            f"Status is not in valid statuses. Got {resp.status}, wants on of {self.valid_statuses} "
                        )
                    return await resp.text()
            except aiohttp.http_exceptions.BadHttpMessage as e:
                raise HealthCheckError("Health test Failed") from e
            except aiohttp.client_exceptions.ClientError as e:
                raise HealthCheckError("Health test failed") from e


@runtime_checkable
class Logger(Protocol):
    """A logger for the deployment."""

    def on_pull(self, log: tuple[str, str]) -> None:
        """When the deployment is pulled, this method is called."""
        ...

    def on_up(self, log: tuple[str, str]) -> None:
        """When the deployment is up, this method is called."""
        ...

    def on_stop(self, log: tuple[str, str]) -> None:
        """When the deployment is stopped, this method is called."""
        ...

    def on_logs(self, log: tuple[str, str]) -> None:
        """When the deployment is logging, this method is called."""
        ...

    def on_down(self, log: tuple[str, str]) -> None:
        """When the deployment is down, this method is called."""
        ...


class Deployment(KoiledModel):
    """A deployment is a set of services that are deployed together."""

    project: Project = Field(default_factory=Project)

    health_checks: List[HealthCheck] = Field(
        default_factory=list,
        description="A list of health checks to run on the deployment. These are run when the deployment is up and running.",
    )
    initialize_on_enter: bool = Field(
        default=False,
        description="Should we initialize the deployment when entering the context manager.",
    )
    inspect_on_enter: bool = Field(
        default=False,
        description="Should we inspect the deployment when entering the context manager.",
    )
    pull_on_enter: bool = Field(
        default=False,
        description="Should we pull the deployment when entering the context manager.",
    )
    up_on_enter: bool = Field(
        default=False,
        description="Should we up the deployment when entering the context manager.",
    )
    health_on_enter: bool = Field(
        default=False,
        description="Should we check the health of the deployment when entering the context manager.",
    )
    down_on_exit: bool = Field(
        default=False,
        description="Should we down the deployment when exiting the context manager.",
    )
    stop_on_exit: bool = Field(
        default=False,
        description="Should we stop the deployment when exiting the context manager.",
    )
    tear_down_on_exit: bool = Field(
        default=False,
        description="Should we tear down the deployment when exiting the context manager.",
    )
    threadpool_workers: int = Field(
        default=10,
        description="The number of workers to use for the threadpool. This is used for the health checks and the log watcher.",
    )

    pull_logs: Optional[List[str]] = Field(
        default=None,
        description="The logs of the pull command. Will be set when the deployment is pulled.",
    )
    up_logs: Optional[List[str]] = Field(
        default=None,
        description="The logs of the up command. Will be set when the deployment is up.",
    )
    stop_logs: Optional[List[str]] = Field(
        default=None,
        description="The logs of the stop command. Will be set when the deployment is stopped.",
    )

    auto_initialize: bool = Field(
        default=True,
        description="Should we automatically initialize the deployment when using it as a context manager.",
    )

    logger: Logger = Field(default_factory=VoidLogger)

    _spec: Optional[ComposeSpec] = None
    _cli: Optional[CLI] = None

    @property
    def spec(self) -> ComposeSpec:
        """A property that returns the compose spec of the deployment.

        THis compose spec can be used to retrieve information about the deployment.
        by inspecting the containers, networks, volumes, etc.

        In the future, this spec will be used to
        retrieve information about the deployment.

        Returns
        -------
        ComposeSpec
            The compose spec.

        Raises
        ------
        NotInspectedError
            If the deployment has not been inspected.
        """
        if self._spec is None:
            raise NotInspectedError(
                "Deployment not inspected. Call await deployment.ainspect() first."
            )
        return self._spec

    async def ainitialize(self) -> "CLI":
        """Initialize the deployment.

        Will initialize the deployment through its project and return the CLI object.
        This method is called automatically when using the deployment as a context manager.

        Returns
        -------
        CLI
           The CLI object.
        """
        self._cli = await self.project.ainititialize()
        return self._cli

    async def aretrieve_cli(self) -> "CLI":
        """Retrieve the CLI object of the deployment."""
        if self._cli is None:
            if self.auto_initialize:
                self._cli = await self.ainitialize()
            else:
                raise NotInitializedError(
                    "Deployment not initialized and auto_initialize is False. Call await deployment.ainitialize() first."
                )

        return self._cli

    async def ainspect(self) -> ComposeSpec:
        """Inspect the deployment.

        Will inspect the deployment through its project and return the compose spec, which
        can be used to retrieve information about the deployment.
        This method is called automatically when using the deployment as a context manager and
        if inspect_on_enter is True.
        Returns
        -------
        ComposeSpec
            The compose spec.

        Raises
        ------
        NotInitializedError
            If the deployment has not been initialized.
        """
        cli = await self.aretrieve_cli()
        self._spec = await cli.ainspect_config()
        return self._spec

    def inspect(self) -> ComposeSpec:
        """Inspect the deployment.

        Will inspect the deployment through its project and return the compose spec, which
        can be used to retrieve information about the deployment.
        This method is called automatically when using the deployment as a context manager and
        if inspect_on_enter is True.

        Returns
        -------
        ComposeSpec
            The compose spec.
        Raises
        ------
        NotInitializedError
            If the deployment has not been initialized.
        """
        return unkoil(self.ainspect)

    def add_health_check(
        self,
        url: Union[str, Callable[[ComposeSpec], str]],
        service: str,
        max_retries: int = 3,
        timeout: int = 10,
        error_with_logs: bool = True,
    ) -> "HealthCheck":
        """Add a health check to the deployment.

        Parameters
        ----------
        url : Union[str, Callable[[ComposeSpec], str]]
            The url to check. Also accepts a function that uses the introspected compose spec to build an url
        service : str
            The service this health check is for.
        max_retries : int, optional
            The maximum retries before the healtch checks fails, by default 3
        timeout : int, optional
            The timeout between retries, by default 10
        error_with_logs : bool, optional
            Should we error with the logs of the service (will inspect container logs of the service), by default True

        Returns
        -------
        HealthCheck
            The health check object.
        """

        check = HealthCheck(
            url=url,
            service=service,
            max_retries=max_retries,
            timeout=timeout,
            error_with_logs=error_with_logs,
        )

        self.health_checks.append(check)
        return check

    async def arun_check(self, check: HealthCheck, retry: int = 0) -> None:
        """Run a health check.

        This method will make a request to the given URL and check the response status.
        If the status is not in the valid statuses, an error will be raised.
        Parameters
        ----------
        check : HealthCheck
            The health check to run.
        retry : int
            The number of retries already done.
        """

        if not self._spec:
            self._spec = await self.ainspect()

        if not self._cli:
            self._cli = await self.ainitialize()

        try:
            await check.acheck(self._spec)
        except HealthCheckError as e:
            if retry < check.max_retries:
                await asyncio.sleep(check.timeout)
                await self.arun_check(check, retry=retry + 1)
            else:
                if not check.error_with_logs:
                    raise HealthCheckError(
                        f"Health check failed after {check.max_retries} retries. Logs are disabled."
                    ) from e

                logs = LogRoll()

                async for log in self._cli.astream_docker_logs(
                    services=[check.service]
                ):
                    logs.append(log)

                raise HealthCheckError(
                    f"Health check failed after {check.max_retries} retries. Logs:\n"
                    + "\n".join(i for _, i in logs)
                ) from e

    async def acheck_health(
        self, timeout: int = 3, retry: int = 0, services: Optional[List[str]] = None
    ) -> None:
        """Check the health of the deployment.

        This method will make a request to all the health checks and check the response status
        concurrently.

        If the status is not in the valid statuses, an error will be raised.

        Parameters
        ----------
        timeout : int
            The timeout between retries.
        retry : int
            The number of retries already done.
        services : Optional[List[str]]
            The list of services to check. If None, all services will be checked.
        """

        if services is None:
            services = [
                check.service for check in self.health_checks
            ]  # we check all services

        await asyncio.gather(
            *[
                self.arun_check(check)
                for check in self.health_checks
                if check.service in services
            ]
        )

    def check_health(
        self,
    ) -> None:
        """Check the health of the deployment.

        This method will make a request to all the health checks and check the response status
        concurrently.
        If the status is not in the valid statuses, an error will be raised.
        """
        return unkoil(self.acheck_health)

    def create_watcher(
        self,
        services: Union[List[str], str],
        tail: Optional[int] = None,
        follow: bool = True,
        no_log_prefix: bool = False,
        timestamps: bool = False,
        since: Optional[str] = None,
        until: Optional[str] = None,
        stream: bool = True,
        wait_for_first_log: bool = True,
        wait_for_logs: bool = False,
        wait_for_logs_timeout: int = 10,
        log_function: Optional[LogFunction] = None,
        append_to_traceback: bool = True,
        capture_stdout: bool = True,
        rich_traceback: bool = True,
    ) -> LogWatcher:
        """Get a logswatcher for a service.

        A logswatcher is an object that can be used to watch the logs of a service, as
        they are being streamed. It is an (async) context manager that should be used
        to enclose any code that needs to watch the logs of a service.

        ```python
         with deployment.logswatcher("service"):
            # do something with service logs
            print(requests.get("http://service").text

        ```

        If you want to watch the logs of multiple services, you can pass a list of service names.

            ```python

            watcher = deployment.logswatcher(["service1", "service2"])
            with watcher:
                # do something with service logs
                print(requests.get("http://service1").text
                print(requests.get("http://service2").text

            print(watcher.collected_logs)

            ```

        Parameters
        ----------
        service_name : Union[List[str], str]
            The name of the service(s) to watch the logs for.

        Returns
        -------
        LogWatcher
            The log watcher object.
        """
        if isinstance(services, str):
            services = [services]

        return LogWatcher(
            cli_bearer=self,
            services=services,
            tail=tail,
            follow=follow,
            no_log_prefix=no_log_prefix,
            timestamps=timestamps,
            since=since,
            until=until,
            stream=stream,
            wait_for_first_log=wait_for_first_log,
            wait_for_logs=wait_for_logs,
            wait_for_logs_timeout=wait_for_logs_timeout,
            log_function=log_function,
            append_to_traceback=append_to_traceback,
            capture_stdout=capture_stdout,
            rich_traceback=rich_traceback,
        )

    async def aup(self, detach: bool = True) -> LogRoll:
        """Up the deployment.

        Will call docker-compose up on the deployment.
        This method is called automatically when using the deployment as a context manager and
        if up_on_enter is True.

        Parameters
        ----------
        detach : bool, optional
            Should we run the up command in detached mode, by default True (otherwise you need to
            call it as a task yourself)

        Returns
        -------
        List[str]
            The logs of the up command.
        """
        cli = await self.aretrieve_cli()
        logs = LogRoll()
        async for log in cli.astream_up(detach=detach):
            logs.append(log)
            self.logger.on_up(log)

        return logs

    def up(self, detach: bool = True) -> LogRoll:
        """Up the deployment.

        Will call docker-compose up on the deployment.
        This method is called automatically when using the deployment as a context manager and
        if up_on_enter is True.

        Parameters
        ----------
        detach : bool, optional
            Should we run the up command in detached mode, by default True (otherwise you need to
            call it as a task yourself, which is not recommended in sync code)

        Returns
        -------
        List[str]
            The logs of the up command.
        """

        return unkoil(self.aup, detach=detach)

    async def arestart(
        self,
        services: Union[List[str], str],
        await_health: bool = True,
        await_health_timeout: int = 3,
    ) -> LogRoll:
        """Restarts a service.

        Will call docker-compose restart on the list of services.
        If await_health is True, will await for the health checks of these services to pass.

        Parameters
        ----------
        services : Union[List[str], str], optional
            The list of services to restart, by default None
        await_health : bool, optional
            Should we await for the health checks to pass, by default True
        await_health_timeout : int, optional
            The time to wait for  before checking the health checks (allows the container to
            shutdown), by default 3, is void if await_health is False

        Returns
        -------
        LogRoll
            The logs of the restart command.
        """
        cli = await self.aretrieve_cli()
        if isinstance(services, str):
            services = [services]

        logs = LogRoll()
        async for log in cli.astream_restart(services=services):
            logs.append(log)

        if await_health:
            await asyncio.sleep(await_health_timeout)
            await self.acheck_health(services=services)

        return logs

    def restart(
        self,
        services: Union[List[str], str],
        await_health: bool = True,
        await_health_timeout: int = 3,
    ) -> LogRoll:
        """Restarts a service. (sync)

        Will call docker-compose restart on the list of services.
        If await_health is True, will await for the health checks of these services to pass.

        Parameters
        ----------
        services : Union[List[str], str], optional
            The list of services to restart, by default None
        await_health : bool, optional
            Should we await for the health checks to pass, by default True
        await_health_timeout : int, optional
            The time to wait for  before checking the health checks (allows the container to
            shutdown), by default 3, is void if await_health is False

        Returns
        -------
        List[str]
            The logs of the restart command.
        """
        return unkoil(
            self.arestart,
            services=services,
            await_health=await_health,
            await_health_timeout=await_health_timeout,
        )

    async def apull(self) -> LogRoll:
        """Pull the deployment.

        Will call docker-compose pull on the deployment.
        This method is called automatically when using the deployment as a context manager and
        if pull_on_enter is True.

        Returns
        -------
        List[str]
            The logs of the pull command.

        Raises
        ------
        NotInitializedError
            If the deployment has not been initialized.
        """
        cli = await self.aretrieve_cli()

        logs = LogRoll()
        async for log in cli.astream_pull():
            logs.append(log)

        return logs

    def pull(self) -> LogRoll:
        """Pull the deployment.

        Will call docker-compose pull on the deployment.
        This method is called automatically when using the deployment as a context manager and
        if pull_on_enter is True.
        Returns
        -------
        List[str]
            The logs of the pull command.
        Raises
        ------
        NotInitializedError
            If the deployment has not been initialized.
        """
        return unkoil(self.apull)

    async def adown(self) -> LogRoll:
        """Down the deployment.

        Will call docker-compose down on the deployment.
        This method is called automatically when using the deployment as a context manager and
        if down_on_exit is True.

        Returns
        -------
        List[str]
            The logs of the down command.
        """
        cli = await self.aretrieve_cli()

        logs = LogRoll()
        async for log in cli.astream_down():
            logs.append(log)

        return logs

    async def aremove(self) -> None:
        """Down the deployment.

        Will call docker-compose down on the deployment.
        This method is called automatically when using the deployment as a context manager and
        if down_on_exit is True.

        Returns
        -------
        List[str]
            The logs of the down command.
        """
        cli = await self.aretrieve_cli()

        return await self.project.atear_down(cli)

    def remove(self) -> None:
        """Remove the project

        Returns
        -------
        List[str]
            The logs of the down command.
        """
        return unkoil(self.aremove)

    def down(self) -> LogRoll:
        """Down the deployment.

        Will call docker-compose down on the deployment.
        This method is called automatically when using the deployment as a context manager and
        if down_on_exit is True.

        Returns
        -------
        List[str]
            The logs of the down command.
        """
        return unkoil(self.adown)

    async def astop(self) -> LogRoll:
        """Stop the deployment.

        Will call docker-compose stop on the deployment.
        This method is called automatically when using the deployment as a context manager and
        if stop_on_exit is True.

        Returns
        -------
        List[str]
            The logs of the stop command.
        """
        cli = await self.aretrieve_cli()

        logs = LogRoll()
        async for log in cli.astream_stop():
            logs.append(log)

        return logs

    def stop(self) -> LogRoll:
        """Stop the deployment.

        Will call docker-compose stop on the deployment.
        This method is called automatically when using the deployment as a context manager and
        if stop_on_exit is True.

        Returns
        -------
        List[str]
            The logs of the stop command.
        """
        return unkoil(self.astop)

    async def aget_cli(self) -> CLI:
        """Get the CLI object of the deployment.

        THis is the defining method of a CLI bearer, and will
        be called by any method that needs the CLI object.
        This is an async method because initializing the CLI object
        is an async operation (as it might incure network calls).
        """
        return await self.aretrieve_cli()

    async def __aenter__(self) -> Self:
        """Async enter method for the deployment.

        Will initialize the project, if auto_initialize is True.
        Will inspect the deployment, if inspect_on_enter is True.
        Will call docker-compose up and pull on the deployment, if
        up_on_enter and pull_on_enter are True respectively.

        """
        if self.initialize_on_enter:
            await self.ainitialize()

        if self.inspect_on_enter:
            await self.ainspect()

        if self.pull_on_enter:
            await self.project.abefore_pull()
            await self.apull()

        if self.up_on_enter:
            await self.project.abefore_up()
            await self.aup()

        if self.health_on_enter:
            if self.health_checks:
                await self.acheck_health()

        return self

    async def __aexit__(
        self,
        exc_type: Optional[Type[BaseException]],
        exc_val: Optional[BaseException],
        traceback: Optional[TracebackType],
    ) -> None:
        """Async exit method for the deployment.

        Will call docker-compose down and stop on the deployment, if
        down_on_exit and stop_on_exit are True respectively.
        """
        if self.stop_on_exit:
            await self.project.abefore_stop()
            await self.astop()

        if self.down_on_exit:
            await self.project.abefore_down()
            await self.adown()

        if self.tear_down_on_exit:
            if self._cli:
                await self.project.atear_down(self._cli)

        self._cli = None



from types import TracebackType
from koil.composition import KoiledModel
import asyncio
from typing import Optional, List, Self, Tuple, Type, Union, Generator
from dokker.cli import CLIBearer
from pydantic import Field
from typing import List

from dokker.types import LogFunction


def format_log_watcher_message(watcher: "LogWatcher", exc_val: Optional[BaseException], rich: bool = True) -> str:
    """Formats the log watcher message for the exception."""
    extra_info = map(
        lambda x: x[1] if x[0] == "STDERR" or watcher.capture_stdout else "",
        watcher.collected_logs,
    )
    # Ensure compatibility with different exception types

    extra_info_str = "\n".join(extra_info)
    return f"{str(exc_val)}\n\nDuring the execution Logwatcher captured these logs from the services {watcher.services}:\n{extra_info_str}"


class LogRoll(list[tuple[str, str]]):
    """A class to roll logs from the log watcher."""

    @property
    def stdout_gen(self) -> Generator[str, None, None]:
        """Generator for stdout logs."""
        for log, x in self:
            if log == "STDOUT":
                yield x

    @property
    def stderr_gen(self) -> Generator[str, None, None]:
        """Generator for stderr logs."""
        for log, x in self:
            if log == "STDOUT":
                yield x

    @property
    def stderr_list(self) -> List[str]:
        """List of stderr logs."""
        return list(self.stderr_gen)

    @property
    def stdout_list(self) -> List[str]:
        """List of stdout logs."""
        return list(self.stdout_gen)

    @property
    def stdout(self) -> str:
        """String of stdout logs joined by new lines."""
        return "\n".join(self.stdout_gen)

    @property
    def stderr(self) -> str:
        """String of stderr logs joined by new lines."""
        return "\n".join(self.stderr_gen)


class LogWatcher(KoiledModel):
    """A class to watch logs from a Docker container."""

    cli_bearer: CLIBearer
    tail: Optional[int] = None
    follow: bool = True
    no_log_prefix: bool = False
    timestamps: bool = False
    since: Optional[str] = None
    until: Optional[str] = None
    stream: bool = True
    services: Union[str, List[str]] = []
    wait_for_first_log: bool = True
    wait_for_logs: bool = False
    wait_for_logs_timeout: int = 10
    collected_logs: LogRoll = Field(default_factory=LogRoll)
    log_function: Optional[LogFunction] = None
    append_to_traceback: bool = True
    capture_stdout: bool = True
    rich_traceback: bool = True

    _watch_task: Optional[asyncio.Task[None]] = None
    _just_one_log: Optional[asyncio.Future[bool]] = None

    async def aon_logs(self, log: Tuple[str, str]) -> None:
        """Asynchronous function to handle logs."""
        if self.log_function:
            if asyncio.iscoroutinefunction(self.log_function):
                await self.log_function(log)
            else:
                self.log_function(log)

    async def awatch_logs(self) -> None:
        """Asynchronous function to watch logs."""
        cli = await self.cli_bearer.aget_cli()
        async for logtuple in cli.astream_docker_logs(
            tail=str(self.tail) if self.tail else None,
            follow=self.follow,
            no_log_prefix=self.no_log_prefix,
            timestamps=self.timestamps,
            since=self.since,
            until=self.until,
            services=self.services,
        ):
            if self._just_one_log is not None and not self._just_one_log.done():
                self._just_one_log.set_result(True)
            await self.aon_logs(logtuple)
            self.collected_logs.append(logtuple)

    async def __aenter__(self) -> Self:
        """Asynchronous context manager to enter the log watcher."""
        self.collected_logs = LogRoll()
        self._just_one_log = asyncio.Future()
        self._watch_task = asyncio.create_task(self.awatch_logs())

        if self.wait_for_first_log:
            await self._just_one_log

        self._just_one_log = asyncio.Future()

        return self

    async def __aexit__(self, exc_type: Optional[Type[BaseException]], exc_val: Optional[BaseException], traceback: Optional[TracebackType]) -> None:
        """Asynchronous context manager to exit the log watcher."""
        if exc_type is not None and self.append_to_traceback:
            new_message = format_log_watcher_message(self, exc_val, rich=self.rich_traceback)
            try:
                new_exc = exc_type(new_message)
            except:  # noqa: E722
                new_exc = Exception(new_message)

            raise new_exc.with_traceback(traceback) from exc_val

        if self.wait_for_logs:
            if self._just_one_log is not None:
                await asyncio.wait_for(self._just_one_log, self.wait_for_logs_timeout)

        if self._watch_task is not None:
            self._watch_task.cancel()

            try:
                await self._watch_task
            except asyncio.CancelledError:
                pass

        self._watch_task = None
